@@ -27,14 +27,14 @@ Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
 and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
 the previous versions.
 
-Modifications Copyright (C) 2008-2018 Vinay Sajip. All rights reserved.
+Modifications Copyright (C) 2008-2019 Vinay Sajip. All rights reserved.
 
 A unittest harness (test_gnupg.py) has also been added.
 """
 
-__version__ = "0.4.3"
+__version__ = "0.4.6"
 __author__ = "Vinay Sajip"
-__date__  = "$13-Jun-2018 12:11:43$"
+__date__  = "$17-Apr-2020 09:35:35$"
 
 try:
     from io import StringIO
@@ -244,6 +244,7 @@ class Verify(object):
         self.sig_timestamp = None
         self.trust_text = None
         self.trust_level = None
+        self.sig_info = {}
 
     def __nonzero__(self):
         return self.valid
@@ -251,41 +252,72 @@ class Verify(object):
     __bool__ = __nonzero__
 
     def handle_status(self, key, value):
+
+        def update_sig_info(**kwargs):
+            sig_id = self.signature_id
+            if sig_id:
+                info = self.sig_info[sig_id]
+                info.update(kwargs)
+
         if key in self.TRUST_LEVELS:
             self.trust_text = key
             self.trust_level = self.TRUST_LEVELS[key]
+            update_sig_info(trust_level=self.trust_level,
+                            trust_text=self.trust_text)
         elif key in ("WARNING", "ERROR"):
             logger.warning('potential problem: %s: %s', key, value)
         elif key == "BADSIG":  # pragma: no cover
             self.valid = False
             self.status = 'signature bad'
             self.key_id, self.username = value.split(None, 1)
+            update_sig_info(keyid=self.key_id, username=self.username,
+                            status=self.status)
         elif key == "ERRSIG":  # pragma: no cover
             self.valid = False
+            parts = value.split()
             (self.key_id,
              algo, hash_algo,
              cls,
-             self.timestamp) = value.split()[:5]
+             self.timestamp) = parts[:5]
+            # Since GnuPG 2.2.7, a fingerprint is tacked on
+            if len(parts) >= 7:
+                self.fingerprint = parts[6]
             self.status = 'signature error'
+            update_sig_info(keyid=self.key_id, timestamp=self.timestamp,
+                            fingerprint=self.fingerprint, status=self.status)
         elif key == "EXPSIG":  # pragma: no cover
             self.valid = False
             self.status = 'signature expired'
             self.key_id, self.username = value.split(None, 1)
+            update_sig_info(keyid=self.key_id, username=self.username,
+                            status=self.status)
         elif key == "GOODSIG":
             self.valid = True
             self.status = 'signature good'
             self.key_id, self.username = value.split(None, 1)
+            update_sig_info(keyid=self.key_id, username=self.username,
+                            status=self.status)
         elif key == "VALIDSIG":
+            fingerprint, creation_date, sig_ts, expire_ts = value.split()[:4]
             (self.fingerprint,
              self.creation_date,
              self.sig_timestamp,
-             self.expire_timestamp) = value.split()[:4]
+             self.expire_timestamp) = (fingerprint, creation_date, sig_ts,
+                                       expire_ts)
             # may be different if signature is made with a subkey
             self.pubkey_fingerprint = value.split()[-1]
             self.status = 'signature valid'
+            update_sig_info(fingerprint=fingerprint, creation_date=creation_date,
+                            timestamp=sig_ts, expiry=expire_ts,
+                            pubkey_fingerprint=self.pubkey_fingerprint,
+                            status=self.status)
         elif key == "SIG_ID":
+            sig_id, creation_date, timestamp = value.split()
+            self.sig_info[sig_id] = {'creation_date': creation_date,
+                                     'timestamp': timestamp}
             (self.signature_id,
-             self.creation_date, self.timestamp) = value.split()
+             self.creation_date, self.timestamp) = (sig_id, creation_date,
+                                                    timestamp)
         elif key == "DECRYPTION_FAILED":  # pragma: no cover
             self.valid = False
             self.key_id = value
@@ -303,6 +335,7 @@ class Verify(object):
             else:
                 self.key_status = 'signing key was revoked'
             self.status = self.key_status
+            update_sig_info(status=self.status, keyid=self.key_id)
         elif key in ("UNEXPECTED", "FAILURE"):  # pragma: no cover
             self.valid = False
             self.key_id = value
@@ -316,17 +349,20 @@ class Verify(object):
                 # write to "/etc/foo" as a non-root user, a "permission denied"
                 # error will be sent as a non-status message.
                 message = 'error - %s' % value
-                parts = value.split()
-                if parts[-1].isdigit():
-                    code = int(parts[-1])
-                    system_error = bool(code & 0x8000)
-                    code = code & 0x7FFF
-                    if system_error:
-                        mapping = self.GPG_SYSTEM_ERROR_CODES
+                operation, code = value.rsplit(' ', 1)
+                if code.isdigit():
+                    code = int(code) & 0xFFFFFF  # lose the error source
+                    if self.gpg.error_map and code in self.gpg.error_map:
+                        message = '%s: %s' % (operation, self.gpg.error_map[code])
                     else:
-                        mapping = self.GPG_ERROR_CODES
-                    if code in mapping:
-                        message = mapping[code]
+                        system_error = bool(code & 0x8000)
+                        code = code & 0x7FFF
+                        if system_error:
+                            mapping = self.GPG_SYSTEM_ERROR_CODES
+                        else:
+                            mapping = self.GPG_ERROR_CODES
+                        if code in mapping:
+                            message = '%s: %s' % (operation, mapping[code])
                 if not self.status:
                     self.status = message
         elif key in ("DECRYPTION_INFO", "PLAINTEXT", "PLAINTEXT_LENGTH",
@@ -343,11 +379,10 @@ class ImportResult(object):
             sec_dups not_imported'''.split()
     def __init__(self, gpg):
         self.gpg = gpg
-        self.imported = []
         self.results = []
         self.fingerprints = []
         for result in self.counts:
-            setattr(self, result, None)
+            setattr(self, result, 0)
 
     def __nonzero__(self):
         if self.not_imported: return False
@@ -417,11 +452,11 @@ class ImportResult(object):
             logger.debug('message ignored: %s, %s', key, value)
 
     def summary(self):
-        l = []
-        l.append('%d imported' % self.imported)
+        result = []
+        result.append('%d imported' % self.imported)
         if self.not_imported:  # pragma: no cover
-            l.append('%d not imported' % self.not_imported)
-        return ', '.join(l)
+            result.append('%d not imported' % self.not_imported)
+        return ', '.join(result)
 
 ESCAPE_PATTERN = re.compile(r'\\x([0-9a-f][0-9a-f])', re.I)
 BASIC_ESCAPES = {
@@ -524,7 +559,7 @@ class ListKeys(SearchKeys):
 
     def fpr(self, args):
         fp = args[9]
-        if fp in self.key_map:  # pragma: no cover
+        if fp in self.key_map and self.gpg.check_fingerprint_collisions:  # pragma: no cover
             raise ValueError('Unexpected fingerprint collision: %s' % fp)
         if not self.in_subkey:
             self.curkey['fingerprint'] = fp
@@ -704,6 +739,10 @@ class DeleteResult(object):
     __bool__ = __nonzero__
 
 
+class TrustResult(DeleteResult):
+    pass
+
+
 class Sign(TextHandler):
     "Handle status messages for --sign"
     def __init__(self, gpg):
@@ -746,6 +785,8 @@ HEX_DIGITS_RE = re.compile(r'[0-9a-f]+$', re.I)
 
 class GPG(object):
 
+    error_map = None
+
     decode_errors = 'strict'
 
     result_map = {
@@ -758,6 +799,7 @@ class GPG(object):
         'scan': ScanKeys,
         'search': SearchKeys,
         'sign': Sign,
+        'trust': TrustResult,
         'verify': Verify,
         'export': ExportResult,
     }
@@ -780,6 +822,9 @@ class GPG(object):
         """
         self.gpgbinary = gpgbinary
         self.gnupghome = gnupghome
+        # issue 112: fail if the specified value isn't a directory
+        if gnupghome and not os.path.isdir(gnupghome):
+            raise ValueError('gnupghome should be a directory (it isn\'t): %s' % gnupghome)
         if keyring:
             # Allow passing a string or another iterable. Make it uniformly
             # a list of keyring filenames
@@ -823,6 +868,10 @@ class GPG(object):
             dot = '.'.encode('ascii')
             self.version = tuple([int(s) for s in m.groups()[0].split(dot)])
 
+        # See issue #97. It seems gpg allow duplicate keys in keyrings, so we
+        # can't be too strict.
+        self.check_fingerprint_collisions = False
+
     def make_args(self, args, passphrase):
         """
         Make a list of command line elements for GPG. The value of ``args``
@@ -921,11 +970,15 @@ class GPG(object):
         while True:
             data = stream.read(1024)
             if len(data) == 0:
+                if on_data:
+                    on_data(data)
                 break
             logger.debug("chunk: %r" % data[:256])
-            chunks.append(data)
+            append = True
             if on_data:
-                on_data(data)
+                append = on_data(data) != False
+            if append:
+                chunks.append(data)
         if _py3k:
             # Join using b'' or '', as appropriate
             result.data = type(data)().join(chunks)
@@ -946,7 +999,8 @@ class GPG(object):
         rr.start()
 
         stdout = process.stdout
-        dr = threading.Thread(target=self._read_data, args=(stdout, result, self.on_data))
+        dr = threading.Thread(target=self._read_data, args=(stdout, result,
+                              self.on_data))
         dr.setDaemon(True)
         logger.debug('stdout reader: %r', dr)
         dr.start()
@@ -956,6 +1010,9 @@ class GPG(object):
         if writer is not None:
             writer.join()
         process.wait()
+        rc = process.returncode
+        if rc != 0:
+            logger.warning('gpg returned a non-zero error code: %d', rc)
         if stdin is not None:
             try:
                 stdin.close()
@@ -996,9 +1053,20 @@ class GPG(object):
             args.extend(['--yes'])
         args.extend(['--output', no_quote(output)])
 
+    def is_valid_passphrase(self, passphrase):
+        """
+        Confirm that the passphrase doesn't contain newline-type characters -
+        it is passed in a pipe to gpg, and so not checking could lead to
+        spoofing attacks by passing arbitrary text after passphrase and newline.
+        """
+        return ('\n' not in passphrase and '\r' not in passphrase and
+                '\x00' not in passphrase)
+
     def sign_file(self, file, keyid=None, passphrase=None, clearsign=True,
                   detach=False, binary=False, output=None, extra_args=None):
         """sign file"""
+        if passphrase and not self.is_valid_passphrase(passphrase):
+            raise ValueError('Invalid passphrase')
         logger.debug("sign_file: %s", file)
         if binary:  # pragma: no cover
             args = ['-s']
@@ -1036,7 +1104,8 @@ class GPG(object):
         """Verify the signature on the contents of the string 'data'
 
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
-        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome="keys")
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
+        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> input = gpg.gen_key_input(passphrase='foo')
         >>> key = gpg.gen_key(input)
         >>> assert key
@@ -1097,14 +1166,17 @@ class GPG(object):
     # KEY MANAGEMENT
     #
 
-    def import_keys(self, key_data):
+    def import_keys(self, key_data, extra_args=None):
         """
         Import the key_data into our keyring.
         """
         result = self.result_map['import'](self)
         logger.debug('import_keys: %r', key_data[:256])
         data = _make_binary_stream(key_data, self.encoding)
-        self._handle_io(['--import'], data, result, binary=True)
+        args = ['--import']
+        if extra_args:
+            args.extend(extra_args)
+        self._handle_io(args, data, result, binary=True)
         logger.debug('import_keys result: %r', result.__dict__)
         data.close()
         return result
@@ -1115,10 +1187,11 @@ class GPG(object):
         >>> import shutil
         >>> shutil.rmtree("keys", ignore_errors=True)
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
-        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome="keys")
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
+        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> os.chmod('keys', 0x1C0)
         >>> result = gpg.recv_keys('pgp.mit.edu', '92905378')
-        >>> assert result
+        >>> if 'NO_EXTERNAL_TESTS' not in os.environ: assert result
 
         """
         result = self.result_map['import'](self)
@@ -1159,6 +1232,8 @@ class GPG(object):
         via pinentry, you should specify expect_passphrase=False. (It's only
         checked for GnuPG >= 2.1).
         """
+        if passphrase and not self.is_valid_passphrase(passphrase):
+            raise ValueError('Invalid passphrase')
         which='key'
         if secret:  # pragma: no cover
             if (self.version >= (2, 1) and passphrase is None and
@@ -1171,6 +1246,8 @@ class GPG(object):
         else:
             fingerprints = [no_quote(fingerprints)]
         args = ['--delete-%s' % which]
+        if secret and self.version >= (2, 1):
+            args.insert(0, '--yes')
         args.extend(fingerprints)
         result = self.result_map['delete'](self)
         if not secret or self.version < (2, 1):
@@ -1196,7 +1273,8 @@ class GPG(object):
         via pinentry, you should specify expect_passphrase=False. (It's only
         checked for GnuPG >= 2.1).
         """
-
+        if passphrase and not self.is_valid_passphrase(passphrase):
+            raise ValueError('Invalid passphrase')
         which=''
         if secret:
             which='-secret-key'
@@ -1263,7 +1341,8 @@ class GPG(object):
         >>> import shutil
         >>> shutil.rmtree("keys", ignore_errors=True)
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
-        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome="keys")
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
+        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> input = gpg.gen_key_input(passphrase='foo')
         >>> result = gpg.gen_key(input)
         >>> fp1 = result.fingerprint
@@ -1277,7 +1356,8 @@ class GPG(object):
 
         if sigs:
             which = 'sigs'
-        else:            which='keys'
+        else:
+            which = 'keys'
         if secret:
             which='secret-keys'
         args = ['--list-%s' % which,
@@ -1318,10 +1398,11 @@ class GPG(object):
         >>> import shutil
         >>> shutil.rmtree('keys', ignore_errors=True)
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
         >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> os.chmod('keys', 0x1C0)
         >>> result = gpg.search_keys('<vinay_sajip@hotmail.com>')
-        >>> assert result, 'Failed using default keyserver'
+        >>> if 'NO_EXTERNAL_TESTS' not in os.environ: assert result, 'Failed using default keyserver'
         >>> #keyserver = 'keyserver.ubuntu.com'
         >>> #result = gpg.search_keys('<vinay_sajip@hotmail.com>', keyserver)
         >>> #assert result, 'Failed using keyserver.ubuntu.com'
@@ -1360,7 +1441,8 @@ class GPG(object):
         control input.
 
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
-        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome="keys")
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
+        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> input = gpg.gen_key_input(passphrase='foo')
         >>> result = gpg.gen_key(input)
         >>> assert result
@@ -1385,7 +1467,8 @@ class GPG(object):
             if str(val).strip():    # skip empty strings
                 parms[key] = val
         parms.setdefault('Key-Type','RSA')
-        parms.setdefault('Key-Length',2048)
+        if 'key_curve' not in kwargs:
+            parms.setdefault('Key-Length',2048)
         parms.setdefault('Name-Real', "Autogenerated Key")
         logname = (os.environ.get('LOGNAME') or os.environ.get('USERNAME') or
                    'unspecified')
@@ -1427,6 +1510,8 @@ class GPG(object):
             always_trust=False, passphrase=None,
             armor=True, output=None, symmetric=False, extra_args=None):
         "Encrypt the message read from the file-like object 'file'"
+        if passphrase and not self.is_valid_passphrase(passphrase):
+            raise ValueError('Invalid passphrase')
         args = ['--encrypt']
         if symmetric:
             # can't be False or None - could be True or a cipher algo value
@@ -1467,7 +1552,8 @@ class GPG(object):
         >>> if os.path.exists("keys"):
         ...     shutil.rmtree("keys", ignore_errors=True)
         >>> GPGBINARY = os.environ.get('GPGBINARY', 'gpg')
-        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome="keys")
+        >>> if not os.path.isdir('keys'): os.mkdir('keys')
+        >>> gpg = GPG(gpgbinary=GPGBINARY, gnupghome='keys')
         >>> input = gpg.gen_key_input(name_email='user1@test', passphrase='pp1')
         >>> result = gpg.gen_key(input)
         >>> fp1 = result.fingerprint
@@ -1515,6 +1601,8 @@ class GPG(object):
 
     def decrypt_file(self, file, always_trust=False, passphrase=None,
                      output=None, extra_args=None):
+        if passphrase and not self.is_valid_passphrase(passphrase):
+            raise ValueError('Invalid passphrase')
         args = ["--decrypt"]
         if output:  # write the output to a file with the specified name
             self.set_output_without_confirmation(args, output)
@@ -1547,9 +1635,12 @@ class GPG(object):
             logger.debug('writing ownertrust info: %s', s);
             os.write(fd, s.encode(self.encoding))
             os.close(fd)
-            result = self.result_map['delete'](self)
+            result = self.result_map['trust'](self)
             p = self._open_subprocess(['--import-ownertrust', fn])
             self._collect_output(p, result, stdin=p.stdin)
+            if p.returncode != 0:
+                raise ValueError('gpg returned an error - return code %d' %
+                                 p.returncode)
         finally:
             os.remove(fn)
         return result
